#*******************************************************************************

# Script name: Split Prospecting Area by Year Overlap
# Inputs: Feature dataset generated by Propsecting Area script
# Outputs: Geodatabase with all combinations of year overlap polygon feature
#          feature classes
# Description: This takes prospecting area and creates a geodatabase of feature
#              classes that represent all possible combinations of overlap
#              for the different years
# Date: 06/05/2019
# Author: Anne Kort

#*******************************************************************************

#------------------------ Import modules: arcpy, copy --------------------------


# Display message
arcpy.AddMessage('Importing packages...')
import arcpy
import copy


#------------------------------ Define functions -------------------------------


# Display message
arcpy.AddMessage('Loading script functions...')

# Define a function to split the name of a file from the path
def pathsplit(filePathName):
    """Splits a file path name into the file path and file name.
    str -> list of 2 strings [path, name]"""
    # I could not think of any better way to do this so...
    # To get the name and path for output
    # Split the output file path string by the \ character
    filePathList = filePathName.split('\\')
    # The path name for the file location is defined by all but the
    # Last string in the Outputlist joined back up with \
    filePath = '\\'.join(filePathList[:-1])
    # The file name is defined by the last item in the list
    fileName = filePathList[-1]
    return [filePath, fileName]

# Define a function to get unique values from an attribute field
# I found this one on stackexchange but I will do my best to explain
def unique_values(table, field):
    """Returns a list of the unique values in the attribute table
    arcpy table and field objects -> list of values"""
    # This with statement temporarily sets up a Search Cursor
    # This is a object in arcpy that will look through every row
    # In a table. The field defines the field to look through
    with arcpy.da.SearchCursor(table, field) as cursor:
        # Then go through every row and add the value to a set
        # A set will only accept unique values
        # Return the sorted set converted into a list
        return list(sorted({row[0] for row in cursor}))

# Define a function to create a list of pairs of a given item and a list
def item_pairs(firstItem, aList):
    """Returns a list of unique pairs of the given item
    with every item in the list
    object, list -> list of 2 item lists"""
    # Create empty list to append to
    pairs = []
    # Loop through every item in list
    for item in aList:
        # Check if the given item is a list type
        if type(firstItem) == list:
            # Check if the current item is in the given item
            if item not in firstItem:
                # If not and the given item is a list
                # Append the new item to the given item
                newItem = firstItem + [item]
                # Add this altered list to the new pairs lsit
                pairs.append(newItem)
        # If the given item is not a list
        elif item != firstItem:
            # Append a 2 item list of the given item and new item to pairs
            pairs.append([firstItem, item])
    # Return a copy of the completed pairs list
    return copy.copy(pairs)

# Define a function to create a list of all unique combinations of all items
# in the list
def all_combos(anyList):
    """Returns a list of lists of all the possible unique combinations of values
    in the list
    List - > list of lists"""
    # Create an empty list to append to
    combosList = []
    # Make a list of unique pairs
    # Go through every item in the list
    for item in anyList:
        # Use the item_pairs function to pair the item with all the other
        # Items after it in the list
        pairedItems = item_pairs(item, anyList[anyList.index(item)+1:])
        # Append the paired item list to the final combos list
        combosList = combosList + pairedItems
    return combosList

# Define function to find years between the years of the combination pair
def between_combo(combination, originalList):
    """Creates a list of all the items inbetween the given combination in
    list order.
    2 lists - > 1 list"""
    # Sort the original list
    originalList.sort()
    # Create an empty list to append to
    betweenCombo = []
    # Find the indecies for the items in the combo
    firstIndex = originalList.index(combination[0])
    secondIndex = originalList.index(combination[1])
    # Set as higher or lower index number
    if firstIndex < secondIndex:
        lowIndex = firstIndex
        highIndex = secondIndex
    elif firstIndex > secondIndex:
        highIndex = firstIndex
        lowIndex = secondIndex
    # Loop through all items in the original list
    for item in originalList:
        # If the item is greater than the low index and less than the high
        if originalList.index(item) > lowIndex and originalList.index(item)\
           < highIndex:
            # Add to the not in combo list
            betweenCombo.append(item)
    return betweenCombo

# Define a function to clip away all years not in combo
def erase_no_combo(intersectFile, noCombo):
    """Iterates through every year not in the combination for an intersect
    file and clips away area from those years
    Feature class, list -> edited feature class"""
    # Set the current file to the input file
    currentFile = intersectFile
    # Name next file with same name plus C
    nextFile = intersectFile + "C"
    # Set counter to 1
    counter = 1
    # Loop through all items in no combo
    for item in noCombo:
        # Display message
        arcpy.AddMessage("Erasing " + item + " from " + intersectFile +" .....")
        # Clip the current file with the given item and save it as the next file
        arcpy.Erase_analysis(currentFile, item, nextFile)
        # Delete the current file
        arcpy.Delete_management(currentFile)
        # Set the next file name as the current file
        currentFile = nextFile
        # Name the next file with the current file name and the counter
        nextFile = currentFile + str(counter)
        # Add 1 to the counter
        counter = counter + 1
    # Return the name of the current file
    return currentFile
    

# Define function to intersect any combination of years
def intersect_years(comboSet, yearList):
    """Intersects the prospecting area for a given set of years and only
    the given set of years.
    2 lists -> creation of 1 feature class"""
    # Create a name for the intersect file that starts with I
    intName = "I"
    # Then go through every item in the combo set and add the last 2 digits
    for item in comboSet:
        intName = intName + item[-2:]
    # Display message
    arcpy.AddMessage("Generating "+intName+" .....")
    # Intersect the files in the comboset without keeping attributes
    arcpy.Intersect_analysis(comboSet, intName, 'ONLY_FID')
    # Delete file if there was no overlap
    if int(arcpy.GetCount_management(intName).getOutput(0)) == 0:
        # Delete it from the geodatabase
        arcpy.Delete_management(intName)
        # Display Message
        arcpy.AddMessage("Deleted empty feature class "+intName)
    else:
        # Display message
        arcpy.AddMessage("Dissolving "+intName+" .....")
        # Dissolve the intersected file so it contains only 1 multipart polygon
        disName = "D" + intName
        arcpy.Dissolve_management(intName, disName)
        # Delete the original intersect file
        arcpy.Delete_management(intName)
        # Erase from the dissolved file
        # Get all the years not in the combo
        eraseYears = between_combo(comboSet, yearList)
        # Clip off all the years not in the combo
        finalFile = erase_no_combo(disName, eraseYears)
        # Rename the final file to match the original name
        arcpy.Rename_management(finalFile, intName)
        # Add field for earlier year in combo
        # Display message
        arcpy.AddMessage("Adding year name fields to "+intName+" .....")
        arcpy.AddField_management(intName, 'Year1Name', 'SHORT')
        # Calculate field to first year
        arcpy.CalculateField_management(intName, 'Year1Name',\
                                        '"'+comboSet[0][-4:]+'"',"PYTHON")
        # Add field for later year in combo
        arcpy.AddField_management(intName, 'Year2Name', 'SHORT')
        # Calculate field to second year
        arcpy.CalculateField_management(intName, 'Year2Name',\
                                        '"'+comboSet[1][-4:]+'"',"PYTHON")
        # Add field for year difference
        arcpy.AddField_management(intName, 'YearDifference', 'SHORT')
        # Calculate year difference field
        arcpy.CalculateField_management(intName, 'YearDifference',\
                                        '!Year2Name!-!Year1Name!',"PYTHON")
        

# Define function to find years not in combination
def not_in_combo(aCombination, aList):
    """Creates a list of all the items not in the given combination from the
    original list.
    2 lists - > 1 list"""
    # Create an empty list to append to
    notInCombo = []
    # Loop through all items in the original list
    for item in aList:
        # If the item is not in the combination
        if item not in aCombination:
            # Add to the not in combo list
            notInCombo.append(item)
    return notInCombo

# Define a function to create a field map with a sum
def sum_field_map(fieldMapping, joinfield):
    """Takes a field and adds a sum field map to a field mapping.
    Field mapping object, string -> field map added to field mapping"""
    # Get the index for the given field
    fieldIndex = fieldMapping.findFieldMapIndex(joinfield)
    fieldmap = fieldMapping.getFieldMap(fieldIndex)

    # Set field map merge rule to sum
    fieldmap.mergeRule = 'Sum'
    fieldMapping.replaceFieldMap(fieldIndex, fieldmap)

# Define a function to calculate year 1 counts
#def calc_year1(joinedFile, rowInFile):
    
        

#-------------------------- Get parameters from user ---------------------------


# First parameter is multiple input feature classes of field sites
# These must be point feature classes with a date attribute
fieldSites = arcpy.GetParameter(0)

# Second parameter is the known area that could be prospected (exposed rock)
# This is an optional parameter
possibleArea = arcpy.GetParameter(1)

# Third the folder to send all the output to
folder = arcpy.GetParameterAsText(2)

# Final parameter is the name of the geodatabase to be output
nameGDB = arcpy.GetParameterAsText(3)

#---------------------------- Set up the workspace -----------------------------


# Set the given folder to the workspace
arcpy.env.workspace = folder

# Create a geoDatabase to contain all the output
# Display message
arcpy.AddMessage('Creating ' + nameGDB + '.gdb...')
arcpy.CreateFileGDB_management(folder, nameGDB + '.gdb')

# Set the geoDatabase to the workspace
workspaceName = folder + "\\" + nameGDB +'.gdb'
arcpy.env.workspace = workspaceName
# Display message
arcpy.AddMessage('Workspace set to ' + str(arcpy.env.workspace))


#----------------------- Create the output feature class -----------------------


# Set spatial reference from first input
spaceRef = arcpy.Describe(fieldSites[0]).spatialReference
# Display spatial reference
arcpy.AddMessage('Spatial Reference set to ' + str(spaceRef.name))

# Create a point feature class in the feature dataset for the merged field sites
combineSites = arcpy.CreateFeatureclass_management\
              (workspaceName, nameGDB+'_FS', 'POINT',\
               spatial_reference = spaceRef)

# Display message
arcpy.AddMessage('Empty feature class created for combined sites.')

# Add a date field to prospecting area feature class
arcpy.AddField_management(combineSites, 'DATE', 'DATE')

# Add a year field to prospecting area feature class
arcpy.AddField_management(combineSites, 'YEAR', 'SHORT')


#---------------- Merge field site inputs into a single output------------------


# Create a variable to map date field from input field sites to output
siteMappings = arcpy.FieldMappings()
siteMappings.addTable(combineSites)

# Loop through each input file to append to output
for item in fieldSites:
    
    # Identify the date field
    dateField=arcpy.ListFields(item, field_type='Date')
    dateAttribute=dateField[0]
    
    # Create an empty Field Map object
    fldMap = arcpy.FieldMap()
    # Add the date attribute from the buffer file as the input filed
    fldMap.addInputField(item,str(dateAttribute.name))
    # Create an output field for the date in the Field Map object
    fieldDate=fldMap.outputField
    # Set the field attributes
    fieldDate.name, fieldDate.type = 'DATE', 'DATE'
    # Set the with with the set attributes back to the Field Map
    fldMap.outputField = fieldDate  # Black magic I found online...
    # Add this Field Map object to the Field Mapppings set
    siteMappings.addFieldMap(fldMap)
    
    # Append the field site file to the final output file
    # The inclusion of the Field Mappings object will append the date column
    arcpy.Append_management(item, combineSites, 'NO_TEST', siteMappings)

    # Display name of appended file
    arcpy.AddMessage('Appended ' + str(str(item).split('\\')[-1]) +\
                     ' to combined sites.')


#-------------------- Calculate year field from date field ---------------------


# Set expression for field calculator in VB script
expression = 'DatePart ( "yyyy" , [DATE] )'
# Use field calculator to add year to YEAR field from DATE field
arcpy.CalculateField_management(combineSites, 'YEAR', expression, "VB")
# Display message
arcpy.AddMessage('Year field added to combined sites attributes.')


#--------------------- Buffer and clip combined site file ----------------------


# Buffer by 15 Meters and dissolve by year
# If possible area was not included
if str(possibleArea) == '':
    # Display message
    arcpy.AddMessage('Possible area not included.')
    arcpy.AddMessage('Buffering...')
    # Do buffer analysis and name file ProspectArea
    prospectArea=arcpy.Buffer_analysis(combineSites,nameGDB+'_PA', "15 Meters",\
                      dissolve_option="LIST", dissolve_field='YEAR')

# If possible area was included
else:
    # Display message
    arcpy.AddMessage('Possible area included.')
    arcpy.AddMessage('Buffering...')
    # Do buffer analysis with a temporary file
    arcpy.Buffer_analysis(combineSites, 'tempBuffer', "15 Meters",\
                      dissolve_option="LIST", dissolve_field='YEAR')
    # Display message
    arcpy.AddMessage('Clipping...')
    # Clip buffer with possible area and save as final ProspectArea file
    prospectArea=arcpy.Clip_analysis('tempBuffer', possibleArea, nameGDB+'_PA')
    # Delete temporary buffer file.
    arcpy.Delete_management('tempBuffer')


#----------------------- Split Prospecting Area by Year ------------------------


# Use split by attributes to make each year propsecting area a seperate file
# Display Message
arcpy.AddMessage('Splitting prospecting area by year...')
arcpy.SplitByAttributes_analysis(prospectArea, workspaceName, 'YEAR')


#----------- Create list of all possible combos of prospecting area ------------


# Use unique values function on YEAR field to get a list of years for the
# Prospecting area
paYears = unique_values(prospectArea, 'YEAR')

# Add T to the front of all to match the prospecting area by year names
# Use enumerate function to get index with item in list
# Loop through list
for index, item in enumerate(paYears):
        # Replace the item existing at the index with a string of T + the item
        paYears[index] = "T" + str(item)

# Find all unique combos of these years with the all_combos function
arcpy.AddMessage('Generating all possible year combinations...')
yearCombos = all_combos(paYears)


#------------------------- Intersect all combinations --------------------------


# Loop through all combinations
for item in yearCombos:
    intersect_years(item, paYears)

# Loop through all single years and erase any overlap
for item in paYears:
    # Name files for areas of years with no overlap
    singleYearName = "I"+str(item[-2:])
    # Display message
    arcpy.AddMessage('Generating ' + singleYearName + " ....")
    # Make a copy of the full year area being erased
    arcpy.CopyFeatures_management(item, singleYearName)
    # Get all other years in a list
    otherYears = not_in_combo([item], paYears)
    # Erase all other years from feature class so only areas of no overlap left
    newFilename = erase_no_combo(singleYearName, otherYears)
    # Rename the left over file back to the original single year name
    arcpy.Rename_management(newFilename, singleYearName)
    # Add field for earlier year in combo
    # Display message
    arcpy.AddMessage("Adding year name fields to "+singleYearName+" .....")
    arcpy.AddField_management(singleYearName, 'Year1Name', 'SHORT')
    # Calculate field to first year
    arcpy.CalculateField_management(singleYearName, 'Year1Name',\
                                    item[-4:],"PYTHON")
    # Add field for later year in combo
    arcpy.AddField_management(singleYearName, 'Year2Name', 'SHORT')
    # Add field for year difference
    arcpy.AddField_management(singleYearName, 'YearDifference', 'SHORT')


#-------------------------- Merge all intersections-----------------------------


# Get a list of all intersected file names
# Get list of all files in the geoDatabase
featureClasses = arcpy.ListFeatureClasses()

# Go through full list, delete nonintersect fc's, and add field for filename
arcpy.AddMessage('Cleaning up intersect files...')
# Create an empty list
intersectFeatures = []
# Loop through all feature classes
for fc in featureClasses:
    #If the first letter is I
    if fc[0] == "I" and int(arcpy.GetCount_management(fc).getOutput(0)) > 0:
        # Add the fc to the list of intersect feature classes
        intersectFeatures.append(fc)

# Loop through all the intersect feature classes and add field for combo names
for fc in intersectFeatures:
    # Add a field for the file combo name
    arcpy.AddField_management(fc, 'ComboName', 'TEXT', field_length = 140)
    # Add the name of the file to the new field
    arcpy.CalculateField_management(fc, 'ComboName', '"'+fc+'"')

# Merge the files
arcpy.AddMessage('Merging generated files...')
yearOverlap = arcpy.Merge_management(intersectFeatures, 'OverlapByYear')

# Loop through and delete all individual files
arcpy.AddMessage('Deleting individual files...')
for fc in intersectFeatures:
    arcpy.Delete_management(fc)

# Add a field for field site count for every year to the merged set
for year in paYears:
    # Name field Y then year
    fieldName = 'Y' + year[1:]
    # Add short integer field for counting by site
    arcpy.AddField_management('OverlapByYear', fieldName, 'SHORT')

# Delete the leftover year field
arcpy.DeleteField_management('OverlapByYear','YEAR')


#--------------------------- Prepare field site data ---------------------------


# Create a field for site count for each year in field site feature class
# Get a list of all years in field sites
fsYears = unique_values(combineSites, 'YEAR')

# Add field for each year and
# Calculate each year field with 1 for field site of that year and 0 for others
# Display message
arcpy.AddMessage('Add year fields to field site data...')
# Set the expression for the field calculator
expression = "fromYear(int(!YEAR!))"
# Loop through each year
for year in fsYears:
    # Name field Y then year
    fieldName = 'Y' + str(year)
    # Add short integer field for counting by site
    arcpy.AddField_management(combineSites, fieldName, 'SHORT')
    # Define a codeblock for calculating 0 or 1
    codeblock = """def fromYear(checkYear):
    if checkYear ==""" + str(year) + """:
        return 1
    else:
        return 0"""
    arcpy.CalculateField_management(combineSites, fieldName, expression,\
                                    "PYTHON", codeblock)
    # Display message
    arcpy.AddMessage('Added field for ' + fieldName)


#--------------------- Spatial join field sites to overlap ---------------------


# Set target, join, and output features
targetFeature = yearOverlap
joinFeature = combineSites
outputFeature = 'JoinedOverlap'

# Display message
arcpy.AddMessage('Generating field map for spatial join...')
# Develop field map to sum up points in an area
# Create a field mappings object
sMappings = arcpy.FieldMappings()
# Add the table of the target features
sMappings.addTable(targetFeature)
# Add the table of the join features
sMappings.addTable(joinFeature)
# Delete the YEAR field
yearMap = sMappings.findFieldMapIndex("YEAR")
sMappings.removeFieldMap(yearMap)
dateMap = sMappings.findFieldMapIndex("DATE")
sMappings.removeFieldMap(dateMap)

# Cycle through all year fields and make sum merge rule
for year in fsYears:
    # Get field name
    fieldName = 'Y' + str(year)
    # Use function to switch to sum rule
    sum_field_map(sMappings, fieldName)

# Do spatial join
# Display message
arcpy.AddMessage('Performing spatial join...')
joinOverlap = arcpy.SpatialJoin_analysis(targetFeature, joinFeature,\
                                         outputFeature,\
                                         field_mapping = sMappings)


#---------------------------- Clean up spatial join ----------------------------


# Delete rows with no site data
# Use a cursor to go through all rows in join count column
# Display message
arcpy.AddMessage('Cleaning spatial join...')
with arcpy.da.UpdateCursor(joinOverlap,"Join_Count") as cursor:
    # Go through every row
    for row in cursor:
        # Display message
        arcpy.AddMessage('Checking row...')
        # If the value of the row is 0
        if row[0] == 0:
            # Display message
            arcpy.AddMessage('Deleted row')
            # Use the cursor to delete the row
            cursor.deleteRow()

# Delete the unnecessary field join count and fid
arcpy.AddMessage('Deleting unnecessary field ...')
arcpy.DeleteField_management('JoinedOverlap', 'Join_Count')
arcpy.DeleteField_management('JoinedOverlap', 'TARGET_FID')


#-------------------- Add early and later year count fields --------------------


# Transfer Year counts to Year1Count and Year2Count fields
# Add field for year1
arcpy.AddField_management('JoinedOverlap', 'Year1Count', 'SHORT')
# Add field for year2
arcpy.AddField_management('JoinedOverlap', 'Year2Count', 'SHORT')

# List year field names
yearFieldNames = []
for year in fsYears:
    yearFieldNames.append("Y"+str(year))
# Add counts to early year field
# Create a list of all the fields to be accessed
fullFields1 = ['Year1Name', 'Year1Count'] + yearFieldNames
# Update early year field with update cursor
with arcpy.da.UpdateCursor('JoinedOverlap',fullFields1) as cursor:
    # Go through every row
    for row in cursor:
        # Set value of early year count to given matching year field
        # Row[1] is the year1count field
        # Row[0] is the year1name field
        # Use the value of the year1count field to get the index of the
        # Corresponding year field from the fullFields list
        row[1] = row[fullFields1.index('Y'+str(row[0]))]
        # Use update cursor function to update row to this value
        cursor.updateRow(row)
        
# Add counts to later year field
# Create a list of all the fields to be accessed
fullFields2 = ['Year2Name', 'Year2Count'] + yearFieldNames
# Update later year field with update cursor
with arcpy.da.UpdateCursor('JoinedOverlap',fullFields2) as cursor:
    # Go through every row
    for row in cursor:
        # If the later year name is not null
        if row[0] != None:
            # Set value of early year count to given matching year field
            # Row[1] is the year2count field
            # Row[0] is the year2name field
            # Use the value of the year2count field to get the index of the
            # Corresponding year field from the fullFields list
            row[1] = row[fullFields2.index('Y'+str(row[0]))]
            # Use update cursor function to update row to this value
            cursor.updateRow(row)


